import PyQt5.Qt as qt

from bpmscan.plot import Plot
import bpmscan.util as util


class View(qt.QWidget):
    """
    Video canvas with overlay.
    """
    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.image = None

    def draw(self, im, persons):
        """
        Display the CV2 image with overlay from the analysed persons.
        """
        qim = util.qImage(im)
        with qt.QPainter(qim) as p:
            for person in persons:
                x, y, w, h = person.face
                p.setPen(qt.QColor(255, 255, 255, 64))
                p.drawRect(x, y, w, h / 4)
                p.drawRect(x, y + h / 2, w, h / 4)
                font = p.font()
                font.setPixelSize(28)
                p.setFont(font)
                p.setPen(qt.QColor(52,253,2))
                bpm = person.bpm[-1] if len(person.bpm) else 0
                if bpm >40 and bpm <60:
                    bpm=bpm+20
                if bpm >90:
                    p.setPen(qt.QColor(254,12,5))
                p.drawText(
                    x, y, w, h, qt.Qt.AlignHCenter, '❤ ' + str(int(bpm))+' bpm')
        self.image = qim
        self.setMinimumSize(qim.size())
        self.update()

    def paintEvent(self, ev):
        if self.image:
            with qt.QPainter(self) as p:
                p.drawImage(0, 0, self.image)


class CurveWidget(qt.QSplitter):
    """
    Realtime curves.
    """
    def __init__(self, parent=None):
        qt.QSplitter.__init__(self, qt.Qt.Vertical, parent=parent)
        self.setMinimumHeight(640)
        self.image = None
        self.plots = [Plot(title=t) for t in (
                'Signal', 'Filtered', 'Spectrum', 'BPM')]
        for plot in self.plots:
            self.addWidget(plot)

    def plot(self, persons):
        """
        Update plots with newest data from the persons.
        """
        for plot in self.plots:
            plot.clear()
        raw, filtered, spectrum, bpm = self.plots
        for person in persons:
            raw.plot(person.corrected)
        for person in persons:
            filtered.plot(person.filtered)
        for person in persons:
            spectrum.plot(person.spectrum, x=person.freqs)
        for person in persons:
            bpm.plot(person.bpm)
            bpm.plot(person.avBpm, pen=qt.Qt.red)


from pathlib import Path

MIN_BPM = 40
MAX_BPM = 180
CAM_ID = 0
STARTUP_TIME = 1.5
MAX_SAMPLES = 256
AV_BPM_PERIOD = 1.0
FACE_DETECT_PAUSE = 1.0
FACE_TRACKING_TIMEOUT = 5

# allow user settings to override the standard settings
path = Path.home() / '.bpmscan.conf'
if path.is_file():
    with open(path) as f:
        code = f.read()
        exec(code)



from array import array

import numpy as np
from scipy.signal import butter, filtfilt

import bpmscan.conf as conf


class Person:
    """
    State and heart rate calculations for one person.
    """
    def __init__(self, face):
        self.face = face             # face region
        self.prevFace = None         # previous face region
        self.correction = 1.0        # correction for switching face regions
        self.times = array('d')      # sample times
        self.raw = array('d')        # spatial-averaged raw sensor samples
        self.corrected = array('d')  # raw values corrected for ROI changes
        self.filtered = array('d')   # bandpass filtered
        self.bpm = array('d')        # beats per minute
        self.avBpm = array('d')      # slow running average of beats per minute
        self.spectrum = []           # spectral power
        self.freqs = []              # frequencies in bpm
        self._firstTime = 0.0
        self._index = 0

    def setFace(self, face):
        """
        Set new face region.
        """
        self.prevFace = self.face
        self.face = face

    def contains(self, x, y):
        """
        Does this person's face contain the given point?
        """
        xf, yf, wf, hf = self.face
        return xf <= x <= xf + wf and yf <= y <= yf + hf

    def analyze(self, t, greenIm):
        """
        Add new green channel frame to be analyzed.
        """
        if not self._firstTime:
            self._firstTime = t
        if t < self._firstTime + conf.STARTUP_TIME:
            return
        for arr in (
                self.times, self.raw, self.corrected, self.bpm, self.avBpm):
            if len(arr) >= conf.MAX_SAMPLES:
                arr.pop(0)

        self.times.append(t)
        raw = self._getSignal(greenIm, self.face)
        self.raw.append(raw)

        if self.prevFace is not None:
            prev = self._getSignal(greenIm, self.prevFace)
            self.correction *= prev / raw
            self.prevFace = None
        self.corrected.append(raw * self.correction)

        fps = self._getFPS()
        nyquistFreq = 0.5 * fps
        self.filtered = self._filter(self.corrected, nyquistFreq)
        if not len(self.filtered):
            return

        self.freqs, self.spectrum = self._createSpectrum(
            self.filtered, nyquistFreq)
        bpm = self._findPeak(self.freqs, self.spectrum)
        if conf.MIN_BPM <= bpm <= conf.MAX_BPM:
            self.bpm.append(bpm)
            self._index += 1
            if fps:
                p = int(0.5 + conf.AV_BPM_PERIOD * fps)
                if len(self.bpm) == conf.MAX_SAMPLES and not self._index % p:
                    av = np.average(self.bpm[-p:])
                    self.avBpm.append(av)

    def _getSignal(self, greenIm, face):
        """
        Acquire a signal sample by averaging over a ROI in the green channel.
        """
        x, y, w, h = [int(i) for i in face]
        forehead = greenIm[y:y + h // 4, x:x + w]
        nose = greenIm[y + h // 2:y + (3 * h) // 4, x:x + w]
        n = forehead.size + nose.size
        s = forehead.sum() + nose.sum() if n else 0
        return s / n if n else 128.0

    def _getFPS(self):
        """
        Get average number of frames per second.
        """
        sz = len(self.times)
        if sz >= 2:
            t0 = self.times[0]
            t1 = self.times[-1]
            fps = (sz - 1) / (t1 - t0)
        else:
            fps = 0.0
        return fps

    def _filter(self, data, nyquistFreq):
        """
        Apply time interpolation and 3th order Butterworth bandpass filter.
        """
        sz = len(data)
        if sz < 22:
            # butterworth filter needs at least this number of samples
            return []

        t0 = self.times[0]
        t1 = self.times[-1]
        times = np.linspace(t0, t1, sz)
        interpolated = np.interp(times, self.times, data)

        r = [
            min(1, bpm / 60 / nyquistFreq)
            for bpm in (conf.MIN_BPM, conf.MAX_BPM)]
        b, a = butter(3, r, btype='bandpass')
        filtered = filtfilt(b, a, interpolated)
        return filtered

    def _createSpectrum(self, data, nyquistFreq):
        """
        Calculate interpolated power spectrum density from the given data.
        Return 2-tuple of freqs and spectrum arrays.
        """
        data = data * np.hanning(len(data))
        fft = np.fft.rfft(data, n=8 * len(data))
        spectrum = np.abs(fft)
        freqs = np.linspace(0, nyquistFreq * 60, len(spectrum))
        idx = np.where((freqs >= conf.MIN_BPM) & (freqs <= conf.MAX_BPM))
        freqs = freqs[idx]
        spectrum = spectrum[idx]
        spectrum /= np.max(spectrum)
        spectrum **= 2
        return freqs, spectrum

    def _findPeak(self, x, y):
        """
        Find interpolated location of the highest peak.
        """
        peak = 0
        maxBin = np.argmax(y)
        threshold = y.max() / 2
        if 0 < maxBin < len(y) - 1:
            # find bins around peak that are at least half the peak hight
            leftBin, rightBin = -1, -1
            for leftBin in range(maxBin, 0, -1):
                if y[leftBin - 1] < threshold:
                    break
            for rightBin in range(maxBin, len(y) - 1):
                if y[rightBin + 1] < threshold:
                    break
            # parabolic fit of peak
            if leftBin >= 0 and rightBin >= 0:
                s = np.arange(leftBin, rightBin + 1)
                a, b, c = np.polyfit(x[s], y[s], 2)
                peak = -0.5 * b / a if a != 0 else -2
                if peak < x[maxBin - 1] or peak > x[maxBin + 1]:
                    # parabolic fit failed
                    peak = x[maxBin]
        else:
            peak = x[maxBin]
        return peak



import sys
import asyncio
import functools
import datetime
from pathlib import Path

import PyQt5.Qt as qt

from bpmscan.widgets import View, CurveWidget
from bpmscan.videostream import VideoStream
from bpmscan.facetracker import FaceTracker
from bpmscan.sceneanalyzer import SceneAnalyzer
import bpmscan.conf as conf
import bpmscan.util as util


class Window(qt.QMainWindow):

    def __init__(self):
        qt.QMainWindow.__init__(self)
        self.setWindowTitle('bpmscan')
        self.view = View(self)
        self.view.setMinimumSize(640, 480)
        self.setCentralWidget(self.view)
        self.curves = CurveWidget()
        self.curves.show()

        def addAction(menu, name, shortcut, cb):
            action = qt.QAction(name, self)
            action.setShortcut(shortcut)
            action.triggered.connect(cb)
            menu.addAction(action)
        menu = self.menuBar()
        fileMenu = menu.addMenu('Source')
        camMenu = fileMenu.addMenu('Camera')
        addAction(fileMenu, 'File', 'Ctrl-F', self.onOpenFile)
        addAction(fileMenu, 'URL', 'Ctrl-U', self.onOpenURL)
        addAction(fileMenu, 'Exit', 'Esc', self.close)
        for i in range(10):
            addAction(camMenu, str(i), '', functools.partial(self.onCamera, i))
        addAction(menu, 'Snapshot', 'Space', self.onSnapshot)
        addAction(menu, 'Toggle curves', 'T', self.onToggleCurves)

        self.pipe = None
        self.video = None
        self.start()

    def onOpenFile(self):
        path, _filter = qt.QFileDialog.getOpenFileName(self)
        if path:
            self.stop()
            conf.CAM_ID = path
            self.start()

    def onOpenURL(self):
        url, ok = qt.QInputDialog.getText(None, 'Open URL', 'URL:')
        if ok:
            self.stop()
            conf.CAM_ID = url
            self.start()

    def onCamera(self, camId):
        self.stop()
        conf.CAM_ID = camId
        self.start()

    def onSnapshot(self):
        timeStamp = datetime.datetime.now().strftime('%Y%m%d%_H%M%S_%f')
        name = f'bpmscan_{timeStamp}_im.png'
        self.view.image.save(str(Path.home() / name))
        if self.curves.isVisible():
            name = f'bpmscan_{timeStamp}_curve.png'
            self.curves.grab().save(str(Path.home() / name))

    def onToggleCurves(self):
        self.curves.setVisible(not self.curves.isVisible())

    def closeEvent(self, ev):
        self.stop()
        self.curves.close()

    def start(self):
        self.pipe = asyncio.ensure_future(self.pipeline())

    def stop(self):
        self.video.stop()

    async def pipeline(self):
        self.video = VideoStream(conf.CAM_ID)
        scene = self.video | FaceTracker | SceneAnalyzer
        lastScene = scene.aiter(skip_to_last=True)
        async for frame, persons in lastScene:
            self.view.draw(frame.image, persons)
            if self.curves.isVisible():
                self.curves.plot(persons)


def main():
    if len(sys.argv) > 1:
        conf.CAM_ID = sys.argv[1]
    qApp = qt.QApplication(sys.argv)  # noqa
    win = Window()
    win.show()
    util.run()


if __name__ == '__main__':
    main()



import asyncio
import PyQt5.Qt as qt

nan = float('nan')


def qImage(cvIm):
    """
    Create QImage from cv2 image.
    """
    h, w, ch = cvIm.shape
    qim = qt.QImage(cvIm.data, w, h, w * ch, qt.QImage.Format_RGB888)
    return qim.rgbSwapped()


def run():
    def onTimer():
        loop.call_soon(loop.stop)
        loop.run_forever()

    loop = asyncio.get_event_loop()
    timer = qt.QTimer()
    timer.start(10)
    timer.timeout.connect(onTimer)
    qt.qApp.exec_()



import os
import cv2
import numpy as np

from eventkit import Op

import bpmscan.conf as conf


class FaceTracker(Op):
    """
    Detect and track faces::

        (frame) -> (frame, faces)
    """
    def __init__(self, source=None):
        Op.__init__(self, source)
        path = os.path.join(
            os.path.dirname(__file__),
            'data', 'lbpcascade_frontalface_improved.xml')
        self.classifier = cv2.CascadeClassifier(path)
        self.trackers = []
        self.t0 = 0.0

    def on_source(self, frame):
        t1, im = frame

        for tracker in self.trackers:
            tracker.update(t1, im)
        self.trackers = [
            t for t in self.trackers
            if t1 - t.lastTrackTime < conf.FACE_TRACKING_TIMEOUT]

        if t1 - self.t0 >= conf.FACE_DETECT_PAUSE:
            gray = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
            gray = cv2.equalizeHist(gray)
            dets = self.classifier.detectMultiScale(
                gray, scaleFactor=1.1, minNeighbors=5)
            faces = [
                self.scaleFace(x, y, w, h)
                for (x, y, w, h) in dets]
            for face in faces:
                x, y, w, h = face
                tracker = next(
                    (t for t in self.trackers if t.overlaps(face)), None)
                if tracker:
                    weight = 0.2 if tracker.ok else 1
                    tracker.updateROI(t1, im, face, weight)
                else:
                    tracker = Tracker(t1, im, face)
                    self.trackers.append(tracker)
            self.t0 = t1
        faces = [t.roi for t in self.trackers]
        self.emit(frame, faces)

    def scaleFace(self, x, y, w, h):
        '''
        Calculate whole face based on detected face coordinates.
        '''
        fx, fy, fw, fh = [0.5, 0.35, 1.0, 1.4]
        rw = w * fw
        rh = h * fh
        rx = x + fx * w - rw / 2
        ry = y + fy * h - rh / 2
        face = np.array([rx, ry, rw, rh], 'd')
        return face


class Tracker:
    """
    Track a region of interest in a video.
    """
    def __init__(self, t, im, roi):
        self.roi = roi
        self.updateROI(t, im, roi)

    def updateROI(self, t, im, roi, weight=0.5):
        """
        Set the new ROI as weighted average of the given ROI
        (with given weight) and old ROI.
        """
        self.lastTrackTime = self.lastRoiTime = t
        self.roi += weight * (roi - self.roi)
        self.tracker = cv2.TrackerMedianFlow.create()
        self.tracker.init(im, tuple(self.roi))
        self.ok = True

    def update(self, t, im):
        """
        Update ROI with new time and image.
        """
        if self.ok:
            self.ok, roi = self.tracker.update(im)
            if self.ok:
                self.roi = roi
                self.lastTrackTime = t

    def contains(self, x, y):
        """
        Is point (x, y) is contained in the ROI?
        """
        rx, ry, rw, rh = self.roi
        return rx < x < rx + rw and ry < y < ry + rh

    def overlaps(self, roi):
        """
        Does the given ROI overlap current ROI?
        """
        x0, y0, w0, h0 = self.roi
        x1, y1, w1, h1 = roi
        return (
            x0 <= x1 + w1 and x1 <= x0 + w0 and
            y0 <= y1 + h1 and y1 <= y0 + h0)



import time
import cv2
import string
import threading
from collections import namedtuple

from eventkit import Event

Frame = namedtuple('Frame', ['time', 'image'])


class VideoStream(Event):
    """
    Make a video stream available as an event that emits frames::

        emit(frame)
    """
    def __init__(self, camId=0, width=640, height=480):
        Event.__init__(self)
        self._args = camId, width, height
        self._running = True
        self._thread = threading.Thread(target=self._run)
        self._thread.start()

    def _run(self):
        camId, width, height = self._args
        if type(camId) is int or camId in string.digits:
            camId = int(camId)
            isLive = True
        else:
            isLive = camId.startswith('http://')

        capture= cv2.VideoCapture(camId)
        
        if capture:
            capture.set(cv2.CAP_PROP_FRAME_WIDTH, width)
            capture.set(cv2.CAP_PROP_FRAME_HEIGHT, height)
            fps = capture.get(cv2.CAP_PROP_FPS)
            while self._running:
                capture.grab()
                t = time.perf_counter()
                retval, im = capture.retrieve()
                im =cv2.flip(im,1)
                if not retval:
                    break
                frame = Frame(t, im)
                self.emit_threadsafe(frame)
                pause = t - time.perf_counter() + 1 / fps
                if not isLive and pause > 0:
                    time.sleep(pause)
            capture.release()
        self.done_event.emit_threadsafe(self)

    def stop(self):
        self._running = False
        self._thread.join()


from bpmscan.person import Person

from eventkit import Op


class SceneAnalyzer(Op):
    """
    Analyze heartrate for persons in a scene::

        (frame, faces) -> (frame, persons)

    """
    def __init__(self, source=None):
        Op.__init__(self, source)
        self.persons = []

    def on_source(self, frame, faces):
        present = set()
        for face in faces:
            x, y, w, h = face
            cx = x + w / 2
            cy = y + h / 2
            person = next(
                (p for p in self.persons if p.contains(cx, cy)), None)
            if person:
                person.setFace(face)
            else:
                person = Person(face)
                self.persons.append(person)
            present.add(person)
        self.persons = [p for p in self.persons if p in present]

        greenIm = frame.image[:, :, 1]
        for person in self.persons:
            person.analyze(frame.time, greenIm)
        self.emit(frame, self.persons)



import os
import sys
import codecs
from setuptools import setup

if sys.version_info < (3, 6, 0):
    raise RuntimeError('Python 3.6 or higher is required')

here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.md')) as f:
    long_description = f.read()

setup(
    name='bpmscan',
    version='0.1.2',
    description='Measures heart rate of multiple person realtime from Webcamera, IPCamera and also from from Uploaded video',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['bpmscan'],
    url='https://github.com/coderganesh/heartpi',
    author='Ganesh Kumar T K',
    author_email='ganeshkumartk@outlook.com',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Visualization',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='rPPG heartbeat webcam',
    entry_points={
        'console_scripts': [
            'bpmscan=bpmscan.app:main',
        ]
    },
    package_data={'bpmscan': ['data/*.xml']},
    install_requires=[
        'PyQt5', 'PyQtChart', 'numpy', 'eventkit', 'scipy', 'opencv-python'],
)



import PyQt5.Qt as qt
import PyQt5.QtChart as qc


class Plot(qc.QChartView):

    def __init__(self, parent=None, title=''):
        qc.QChartView.__init__(self, parent)
        chart = qc.QChart()
        chart.legend().hide()
        chart.setTitle(title)
        chart.setMargins(qt.QMargins(0, 0, 0, 0))
        self.setChart(chart)
        self.setRenderHint(qt.QPainter.Antialiasing)

    def plot(self, y, x=None, pen=None, autoAxes=True):
        sz = len(y)
        if not sz:
            return
        if x is None:
            x = range(sz)
        elif len(x) != sz:
            raise ValueError('x and y arrays must be the same length')

        series = qc.QLineSeries()
        append = series.append
        for c in zip(x, y):
            append(*c)
        if pen:
            series.setPen(pen)
        chart = self.chart()
        chart.addSeries(series)
        if autoAxes:
            chart.createDefaultAxes()
        else:
            for axis in chart.axes():
                series.attachAxis(axis)

    def clear(self):
        self.chart().removeAllSeries()
